#############################################################################
# disasm.tcl Copyright 2004 r0lZ
# Disassemble a DVD command to human readeable form.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################


namespace eval ::disasm {

namespace export disasm cmdname dump_all_commands formatCommand

set SPRMlist {
	{preferred menu language}
	{audio stream number}
	{sub-picture stream number}
	{angle number}
	{title number in volume}
	{title number in VTS}
	{PGC number}
	{Chapter number}
	{highlighted button number}
	{navigation timer}
	{PGC jump for nav timer}
	{Karaoke audio mixing mode}
	{parental management country code}
	{parental level}
	{video preference and current mode}
	{player audio capabilities}
	{preferred audio language}
	{preferred audio language extension}
	{preferred sub-picture language}
	{preferred sub-picture language extension}
	{player region code}
	{reserved}
	{reserved}
	{reserved for extended playback mode}
}

set link1list {
	LinkNoLink
	LinkTopCell
	LinkNextCell
	LinkPrevCell
	{<ILLEGAL LINK COMMAND>}
	LinkTopPG
	LinkNextPG
	LinkPrevPG
	{<ILLEGAL LINK COMMAND>}
	LinkTopPGC
	LinkNextPGC
	LinkPrevPGC
	LinkGoUpPGC
	LinkTailPGC
	{<ILLEGAL LINK COMMAND>}
	{<ILLEGAL LINK COMMAND>}
	RSM
}

set vtsmenulist {
	-
	-
	-
	Root
	Sub-picture
	Audio
	Angle
	Chapter
}

set set_operands_list {
	{<ILLEGAL OPERATOR 0>}
	{=(mov)}
	{<->(swap)}
	{+=(add)}
	{-=(sub)}
	{*=(mul)}
	{/=(div)}
	{%=(mod)}
	{?=(rnd)}
	{&=(and)}
	{|=(or)}
	{^=(xor)}
	{<ILLEGAL OPERATOR 12>}
	{<ILLEGAL OPERATOR 13>}
	{<ILLEGAL OPERATOR 14>}
	{<ILLEGAL OPERATOR 15>}
}

set set_operands_list {
	{<ILLEGAL OPERATOR 0>}
	{=(mov)}
	{<->(swap)}
	{+=(add)}
	{-=(sub)}
	{*=(mul)}
	{/=(div)}
	{%=(mod)}
	{?=(rnd)}
	{&=(and)}
	{|=(or)}
	{^=(xor)}
	{<ILLEGAL OPERATOR 12>}
	{<ILLEGAL OPERATOR 13>}
	{<ILLEGAL OPERATOR 14>}
	{<ILLEGAL OPERATOR 15>}
}

#############################################################################
## Return the most or less significant nibble of a byte of the hex in hexadecimal.
## nibble: 0 for most significant nibble, 1 for less

proc get_nibble_hex {hex bytenum nibble} {
	return [string range [lindex $hex $bytenum] $nibble $nibble]
}

#############################################################################
## Return the most or less significant nibble of a byte of the hex command in
## decimal form.
## nibble: 0 for most significant nibble, 1 for less

proc get_nibble {hex bytenum nibble} {
	return [expr "0X[string range [lindex $hex $bytenum] $nibble $nibble]"]
}

#############################################################################
## Return the string value of the most or less significant nibble of a byte
## of the hex command, with range validation.
## nibble: 0 for most significant nibble, 1 for less

proc get_nibble_val {hex bytenum nibble min max} {
	set v [get_nibble $hex $bytenum $nibble]
	if {$v < $min || $v > $max} {
		return "<ILLEGAL VALUE $v>"
	} else {
		return $v
	}
}

#############################################################################
## Return a byte of the hex command in decimal form.

proc get_byte {hex bytenum} {
	return [expr "0X[lindex $hex $bytenum]"]
}

#############################################################################
## Return the string value of a byte of the hex command, with range validation.

proc get_byte_val {hex bytenum min max} {
	set v [expr "0X[lindex $hex $bytenum]"]
	if {$v < $min || $v > $max} {
		return "<ILLEGAL VALUE $v>"
	} else {
		return $v
	}
}

#############################################################################
## Return a word (2 bytes) of the hex command in decimal form.
## The argument 'bytenum' is the index of the first byte of the word (0-6)
## If 'showascii' is true, then append the ascii value of the word.

proc get_word {hex bytenum showascii} {
	set r [expr "0X[lindex $hex $bytenum][lindex $hex [expr $bytenum + 1]]"]
	if $showascii {
		set ascii ""
		set num $r
		while {$num > 0} {
			set c [expr $num % 256]
			set num [expr $num / 256]
			if {$c <= 0X1F || ($c >= 0X80 && $c <= 0X9F)} {
				set ascii ".$ascii"
			} else {
				set ascii "[format {%c} $c]$ascii"
			}
		}
		while {[string length $ascii] < 2} {
			set ascii ".$ascii"
		}
		set r "$r \"${ascii}\""
	}
	return $r
}

#############################################################################
## Return the string value of a word (2 bytes) of the hex command, with range
## validation.
## The argument 'bytenum' is the index of the first byte of the word (0-6)

proc get_word_val {hex bytenum min max} {
	set v [get_word $hex $bytenum false]
	if {$v < $min || $v > $max} {
		return "<ILLEGAL VALUE $v>"
	} else {
		return $v
	}
}

#############################################################################
## Return the GPRM number in the form "gprm(X)" or "gprm(ILLEGAL GPRM X)"

proc operand_gprm {gprm} {
	if {$gprm > 15} {
		return "gprm\(ILLEGAL GPRM $gprm\)"
	}
	return "gprm\($gprm\)"
}

#############################################################################
## Return the SPRM number and name in the form "sprm(X:sprm name)" or
## "sprm(ILLEGAL SPRM X)"

proc operand_sprm {sprm} {
	set sprm [expr $sprm & 0X7F]
	if {$sprm > 23} {
		return "sprm\(ILLEGAL SPRM $sprm\)"
	}
	return "sprm\($sprm:[lindex $::disasm::SPRMlist $sprm]\)"
}

#############################################################################
## Return a register (1 byte, either GPRM or SPRM) of the hex command,
## in the form "gprm(X)", "sprm(X:sprm description)", "gprm(ILLEGAL GPRM X)"
## or "sprm(ILLEGAL SPRM X)"

proc get_register {hex bytenum} {
	set byte [expr 0X[lindex $hex $bytenum]]
	if {[expr $byte & 0X80] == 0} {
		return [operand_gprm $byte]
	} else {
		return [operand_sprm $byte]
	}
}

#############################################################################
## Return the IF operator as text.  Accept both Value or GPRM/SPRM operators.

proc if_operator_text {if_operand_num} {
	return [lindex {{<ILLEGAL OPERATOR>} & == != >= > <= <} [expr $if_operand_num & 7]]
}

#############################################################################
## Return the VTSM menu name or "<ILLEGAL MENU #X>" for illegal menu

proc get_vtsm_menu {hex} {
	set menunum [get_nibble $hex 5 1]
	if {[catch {set txt [lindex $::disasm::vtsmenulist $menunum]}] > 0} {
		return "<ILLEGAL MENU #$menunum>"
	} else {
		if {$txt == "-"} {
			return "<ILLEGAL MENU #$menunum>"
		} else {
			return "$txt menu"
		}
	}
}


#############################################################################
## Return the IF part of the command as text
## This is for IF type 1 (with args in bytes 3, 4 and 5, value allowed)

proc if1 {hex} {
	set iftype [get_nibble $hex 1 0]
	if {[expr $iftype & 7] == 0} {
		set txt ""
	} else {
		set txt "if ( [get_register $hex 3]"
		set txt "$txt [if_operator_text $iftype]"
		if {[expr $iftype & 8] == 0} {
			set txt "$txt [get_register $hex 5]"
		} else {
			set txt "$txt [get_word $hex 4 true]"
		}
		set txt "$txt ) then \{ "
	}
	return $txt
}

#############################################################################
## Return the IF part of the command as text
## This is for IF type 2 (with args in bytes 6 and 7, value not allowed)

proc if2 {hex} {
	set iftype [get_nibble $hex 1 0]
	if {[expr $iftype & 7] == 0} {
		set txt ""
	} else {
		set txt "if ( [get_register $hex 6]"
		set txt "$txt [if_operator_text $iftype]"
		set txt "$txt [get_register $hex 7]"
		set txt "$txt ) then \{ "
	}
	return $txt
}

#############################################################################
## Return a Link command, without the button number.
## Do not check if it is legal to have a link command with the current command.
## Type: 1: allow full range of link commands
##       2: allow only link commands with K=1. K is not present in the command,
##          and assumed to be 1 (used for SetCLnk, CSetCLnk or CSetLnk).
## Return the link command and arguments (without the highlight button),
##        or "-" for error.

proc link {hex type} {
	if {$type == 1} {
		set k [get_nibble $hex 1 1]
	} else {
		set k 1
	}

	if {$k == 0} {
		set txt ""
	} else {
		switch $k {
			1 {
				# Link* (k=1)
				set ll [get_byte $hex 7]
				if {[catch {set txt [lindex $::disasm::link1list $ll]}] != 0} {
						set txt "ILLEGAL LINK COMMAND \[20 01 00 00 00 00 00 >[format %02X $ll]<\]"
				} else {
					set txt "$txt"
				}
			}
			4 {
				# LinkPGCN
				set txt "LinkPGCN PGC [get_word_val $hex 6 1 65535]"
			}
			5 {
				# LinkPTT
				set chapter [expr [get_word $hex 6 false] & 0X03FF]
				if {$chapter == 0 || $chapter > 999} {
					set chapter "<ILLEGAL PTT $chapter>"
				}
				set txt "LinkPTT Chapter $chapter"
			}
			6 {
				# LinkPGN
				set txt "LinkPGN Program [get_byte_val $hex 7 1 99]"
			}
			7 {
				# LinkCN
				set txt "LinkCN Cell [get_byte_val $hex 7 1 255]"
			}
			default {
				set txt "ILLEGAL LINK COMMAND \[20 0>[format %X $k]< 00 00 00 00 00 00\]"
			}
		}
	}
	return $txt
}

#############################################################################
## Return ", button #" (number of the highlighted button) as text, or the empty
## string if button=0.  Check for illegal value.

proc get_hbutton_val {hex} {
	set val [expr [get_byte $hex 6] & 0XFC]
	set but [expr $val / 4]
	set val [expr $val * 0X100]
	if {$but == 0} {return ""}
	if {$but > 36} {
		return ", ILLEGAL BUTTON $but ($val)"
	}
	return ", button $but ($val)"
}

#############################################################################
## Return the text for a Set, SetCLnk, CSetLnk and CSetCLnk operation.
## Check for illegal usage of the operand with the operator 2.
## args: op1, operator, op2: values, direct (1=direct mode, 0=indirect)

proc get_set_operation_text {op1 operator op2 direct} {

	# build start of string with operand 1
	set txt "Set [operand_gprm $op1]"

	# build default operator string
	set operatortxt [lindex $::disasm::set_operands_list $operator]

	# check for illegal usage of operator with operand 2
	if {$direct} {
		set op2txt $op2

		# direct mode is incompatible with SWAP operator
		if {$operator == 2} {
			set operatortxt "<ILLEGAL OPERATOR [string toupper $operatortxt] FOR VALUE OPERAND>"
		}
	} else {
		if {[expr $op2 & 0X80] == 0} {
			set op2txt [operand_gprm $op2]
		} else {
			set op2txt [operand_sprm $op2]

			# SPRM indirect mode is incompatible with most operators
			if {$operator!=1 && $operator!=9 && $operator!=10 && $operator!=11} {
				set operatortxt "<ILLEGAL OPERATOR [string toupper $operatortxt] FOR SPRM OPERAND>"
			}
		}
	}
	set txt "$txt $operatortxt $op2txt"
	return $txt
}

#############################################################################
## Parse DVD hex command, and return a string in human readeable form

proc disasm {hex} {
	set cmdtype [get_nibble $hex 0 0]
	set txt ""

	switch $cmdtype {
		0 {
			# special/flow control commands
			set iftxt [if1 $hex]

			switch [::disasm::get_nibble $hex 1 1] {
				0 {
					# NOP
					set txt "${iftxt}NOP"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				1 {
					# Goto
					set txt "${iftxt}Goto line [::disasm::get_byte_val $hex 7 1 128]"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				2 {
					# Break
					set txt "${iftxt}Break"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				3 {
					# SetTmpPML
					set txt "${iftxt}(SetTmpPML) Set parental level =(mov) "
					if {[::disasm::get_nibble $hex 6 1] == 15} {
						set txt "${txt}15(none)"
						set linenum [::disasm::get_byte $hex 7]
						if {$linenum > 0} {
							set txt "${txt}, Goto line [::disasm::get_byte_val $hex 7 1 128]"
						}
					} else {
						set txt "${txt}[::disasm::get_nibble_val $hex 6 1 1 8]"
						set txt "${txt}, goto line [::disasm::get_byte_val $hex 7 1 128]"
					}
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				default {
					return "ILLEGAL FLOW CONTROL COMMAND \[0>[::disasm::get_nibble_hex $hex 1 1]< 00 00 00 00 00 00 00\]"
				}
			}
		}
		2 {
			# link commands
			set iftxt [if1 $hex]

			set linktxt [link $hex 1]
			if {[string first "ILLEGAL" $linktxt] >= 0} {return $linktxt}
			set linkcmd [lindex $linktxt 0]

			if {[::disasm::get_nibble $hex 1 1] == 4} {
				set txt "${iftxt}${linktxt}"	;# LinkPGCN: no button
			} else {
				set txt "${iftxt}${linktxt}[::disasm::get_hbutton_val $hex]"
			}
			if {$iftxt != ""} {set txt "$txt \}"}
		}
		3 {
			# jump/call commands
			set iftxt [if2 $hex]

			switch [::disasm::get_nibble $hex 1 1] {
				1 {
					# Exit
					set txt "${iftxt}Exit"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				2 {
					# JumpTT
					set txt "${iftxt}(JumpTT) Jump to Title [::disasm::get_byte_val $hex 5 1 99]"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				3 {
					# JumpVTS_TT
					set txt "${iftxt}(JumpVTS_TT) Jump to TTN [::disasm::get_byte_val $hex 5 1 99] in this VTS"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				5 {
					# JumpVTS_PTT
					set txt "${iftxt}(JumpVTS_PTT) Jump to TTN [::disasm::get_byte_val $hex 5 1 99]"
					set txt "$txt, Chapter [get_word_val $hex 2 1 999]"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				6 {
					# JumpSS *
					switch [::disasm::get_nibble $hex 5 0] {
						0 {
							# JumpSS-FP
							set txt "${iftxt}(JumpSS) Jump to First Play PGC"
						}
						4 {
							# JumpSS-VMGM-menu
							set menu [::disasm::get_nibble $hex 5 1]
							if {$menu == 2} {
								set txt "${iftxt}(JumpSS) Jump to VMGM Title menu"
							} else {
								set txt "${iftxt}(JumpSS) Jump to VMGM <ILLEGAL MENU #$menu>"
							}
						}
						8 {
							# JumpSS-VTSM
							set txt "${iftxt}(JumpSS) Jump to"
							set v [::disasm::get_byte_val $hex 4 0 99]
							if {$v == 0} {
								set txt "$txt VTSM of last visited VTS"
							} else {
								set txt "$txt VTSM $v"
							}
							set txt "$txt, [get_vtsm_menu $hex]"
							set txt "$txt (TTN [::disasm::get_byte_val $hex 3 0 99])"
						}
						12 {
							# JumpSS-VMGM-pgc
							set txt "${iftxt}(JumpSS) Jump to VMGM PGC [get_word_val $hex 2 1 999]"
						}
						default {
							return "ILLEGAL JumpSS COMMAND \[30 06 00 00 00 >[::disasm::get_nibble_hex $hex 5 0]<0 00 00\]"
						}
					}
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				8 {
					# CallSS *
					set cell [::disasm::get_byte $hex 4]
					if {$cell > 0} {
						set celltxt ", resume cell [::disasm::get_byte_val $hex 4 1 255]"
					} else {
						set celltxt ""
					}

					switch [::disasm::get_nibble $hex 5 0] {
						0 {
							# CallSS-FP
							set txt "${iftxt}(CallSS) Call the First Play PGC$celltxt"
						}
						4 {
							# CallSS-VMGM-menu
							set menu [::disasm::get_nibble $hex 5 1]
							if {$menu == 2} {
								set txt "${iftxt}(CallSS) Call the VMGM Title menu$celltxt"
							} else {
								set txt "${iftxt}(CallSS) Call the VMGM <ILLEGAL MENU #$menu>$celltxt"
							}
						}
						8 {
							# CallSS-VTSM
							set txt "${iftxt}(CallSS) Call the VTSM [get_vtsm_menu $hex]"
							set txt "$txt of the current TTN$celltxt"
						}
						12 {
							# CallSS-VMGM-pgc
							set txt "${iftxt}(CallSS) Call the VMGM PGC [get_word_val $hex 2 1 999]$celltxt"
						}
						default {
							return "ILLEGAL CallSS COMMAND \[30 08 00 00 00 >[::disasm::get_nibble_hex $hex 5 0]<0 00 00\]"
						}
					}
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				default {
					return "ILLEGAL Jump/Call COMMAND \[30 0>[::disasm::get_nibble_hex $hex 1 1]< 00 00 00 00 00 00\]"
				}
			}
		}
		4 -
		5 {
			# Set SPRM commands

			# ces commandes ne peuvent pas avoir en meme temps un IF et un Link
			set linktype [::disasm::get_nibble $hex 1 1]
			if {[expr [::disasm::get_nibble $hex 1 0] & 7] > 0 && $linktype > 0} {
				return "ILLEGAL USE OF BOTH IF AND LINK \[[format %X $cmdtype]0 >[lindex $hex 1]< 00 00 00 00 00 00\]"
			}
			set iftxt [if2 $hex]
			if {$linktype > 0} {
				set l [link $hex 1]
				if {$l == "" } {
					set linktxt ""
				} else {
					if {$linktype == 4} {
						set linktxt " ; $l"
					} else {
						set linktxt " ; $l[::disasm::get_hbutton_val $hex]"
					}
				}
			} else {
				set linktxt ""
			}

			switch [::disasm::get_nibble $hex 0 1] {
				1 {
					# SetSTN
					set audio [::disasm::get_byte $hex 3]
					if {[expr $audio & 0X80] > 0} {
						if {$cmdtype == 5} {
							set srcval [expr $audio - 0X80]
							if {$srcval == 15} {
								set audiostream "15:none"
							} elseif {$srcval < 8} {
								set audiostream $srcval
							} else {
								set audiostream "<ILLEGAL STREAM $srcval>"
							}
						} else {
							set srcgprm [expr $audio - 0X80]
							if {$srcgprm < 16} {
								set audiostream "gprm\($srcgprm\)"
							} else {
								set audiostream "gprm\(ILLEGAL GPRM $srcgprm\)"
							}
						}
						set audiotxt "Set Audio stream = $audiostream"
					} else {
						set audiotxt ""
					}
					set subpic [::disasm::get_byte $hex 4]
					if {[expr $subpic & 0X80] > 0} {
						set subpiconoff ""
						if {$cmdtype == 5} {
							set onflag [expr $subpic & 0X40]
							set srcval [expr $subpic - 0X80 - $onflag]
							if {$srcval == 62} {
								set subpicstream "62:none"
							} elseif {$srcval == 63} {
								set subpicstream "63:forced"
							} elseif {$srcval < 32} {
								set subpicstream $srcval
								if {$onflag != 0} {
									set subpiconoff ", on"
								} else {
									set subpiconoff ", off"
								}
							} else {
								set subpicstream "<ILLEGAL STREAM $srcval>"
							}
						} else {
							set srcgprm [expr $subpic - 0X80]
							if {$srcgprm < 16} {
								set subpicstream "gprm\($srcgprm\)"
							} else {
								set subpicstream "gprm\(ILLEGAL GPRM $srcgprm\)"
							}
						}
						if {$audiotxt != ""} {set coma " ; "} else {set coma ""}
						set subpictxt "${coma}Set Sub-picture stream = $subpicstream$subpiconoff"
					} else {
						set subpictxt ""
					}

					set angle [::disasm::get_byte $hex 5]
					if {[expr $angle & 0X80] > 0} {
						if {$cmdtype == 5} {
							set srcval [expr $angle - 0X80]
							if {$srcval < 10} {
								set anglestream $srcval
							} else {
								set anglestream "<ILLEGAL ANGLE $srcval>"
							}
						} else {
							set srcgprm [expr $angle - 0X80]
							if {$srcgprm < 16} {
								set anglestream "gprm\($srcgprm\)"
							} else {
								set anglestream "gprm\(ILLEGAL GPRM $srcgprm\)"
							}
						}
						if {$audiotxt != "" || $subpictxt != ""} {set coma " ; "} else {set coma ""}
						set angletxt "${coma}Set Angle = $anglestream"
					} else {
						set angletxt ""
					}

					set txt "${iftxt}(SetSTN) $audiotxt$subpictxt$angletxt$linktxt"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				2 {
					# SetNVTMR
					set txt "${iftxt}(SetNVTMR) Set Navigation Timer =(mov)"
					if {$cmdtype == 5} {
						set txt "$txt [get_word $hex 2 false]"
					} else {
						set txt "$txt [::disasm::get_register $hex 3]"
					}
					set jmp [get_word_val $hex 4 1 999]
					set txt "$txt, Jump to PGC $jmp"

					set txt "$txt$linktxt"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				3 {
					# SetGPRMMD
					set txt "${iftxt}(SetGPRMMD) Set gprm([::disasm::get_nibble $hex 5 1])"
					set mode [expr [::disasm::get_nibble $hex 5 0] & 8]
					if {$mode == 0} {
						set txt "$txt in register mode"
					} elseif {$mode == 8} {
						set txt "$txt in counter mode"
					} else {
						set txt "$txt in <ILLEGAL MODE $mode>"
					}
					if {$cmdtype == 5} {
						set txt "$txt =(mov) [get_word $hex 2 true]"
					} else {
						set txt "$txt =(mov) [::disasm::get_register $hex 3]"
					}
					set txt "$txt$linktxt"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				4 {
					# SetAMXMD
					set txt "${iftxt}(SetAMXMD) Set Karaoke Audio Mixing Mode =(mov)"
					if {$cmdtype == 5} {
						set txt "$txt [get_word_val $hex 4 0 7196]"
					} else {
						set txt "$txt [::disasm::get_register $hex 5]"
					}
					set txt "$txt$linktxt"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				6 {
					# SetHL_BTN
					set txt "${iftxt}(SetHL_BTN) Set Highlighted Button =(mov)"
					if {$cmdtype == 5} {
						set but [expr [::disasm::get_byte $hex 4] / 4]
						set butval [expr $but * 1024]
						if {$but == 0 || $but > 36} {set but "<ILLEGAL BUTTON $but>"}
						set txt "$txt $butval (button $but)"
					} else {
						set txt "$txt [::disasm::get_register $hex 5]"
					}
					# no button here because it is already set!
					if {$linktype > 0} {
						set l [link $hex 1]
						if {$l == "" } {
							set linktxt ""
						} else {
							set linktxt " ; $l"
							set but [expr [::disasm::get_byte $hex 6] / 4]
							if {$but > 0} {
								set linktxt "$linktxt, <REDEFINED BUTTON $but>"
							}
						}
					} else {
						set linktxt ""
					}
					set txt "$txt$linktxt"
					if {$iftxt != ""} {set txt "$txt \}"}
				}
				default {
					return "ILLEGAL Set COMMAND \[>[lindex $hex 0]< 00 00 00 00 00 00 00\]"
				}
			}
		}
		6 -
		7 {
			# Set command

			# Set ne peut pas avoir en meme temps un IF et un Link
			set iftype   [::disasm::get_nibble $hex 1 0]
			set linktype [::disasm::get_nibble $hex 1 1]
			if {[expr $iftype & 7] > 0 && $linktype > 0} {
				return "ILLEGAL USE OF BOTH IF AND LINK \[[format %X $cmdtype]0 >[lindex $hex 1]< 00 00 00 00 00 00\]"
			}
			# IF part is different of the other IFs
			if {[expr $iftype & 7] == 0} {
				set iftxt ""
			} else {
				set iftxt "if ( [::disasm::get_register $hex 2]"
				set iftxt "$iftxt [if_operator_text $iftype]"
				if {[expr $iftype & 8] == 0} {
					set iftxt "$iftxt [::disasm::get_register $hex 7]"
				} else {
					set iftxt "$iftxt [get_word $hex 6 true]"
				}
				set iftxt "$iftxt ) then \{ "
			}
			# Link part
			if {$linktype > 0} {
				set l [link $hex 1]
				if {$l == "" } {
					set linktxt ""
				} else {
					if {$linktype == 4} {
						set linktxt " ; $l"
					} else {
						set linktxt " ; $l[::disasm::get_hbutton_val $hex]"
					}
				}
			} else {
				set linktxt ""
			}
			# Set part
			set direct [expr $cmdtype & 1]
			if $direct {
				set op2 [get_word $hex 4 true]
			} else {
				set op2 [::disasm::get_byte $hex 5]
			}
			set optxt [get_set_operation_text [::disasm::get_byte $hex 3] [::disasm::get_nibble $hex 0 1] $op2 $direct]
			# concat all parts
			set txt "$iftxt$optxt$linktxt"
			if {$iftxt != ""} {set txt "$txt \}"}
		}
		8 -
		9 {
			# SetCLnk

			# The IF operand 1 GPRM and Set destination GPRM are shared
			set sharedreg [::disasm::get_nibble $hex 1 1]

			set iftype [::disasm::get_nibble $hex 1 0]
			if {[expr $iftype & 7] == 0} {
				set iftxt ""
			} else {
				set iftxt " if ( [operand_gprm $sharedreg]"
				set iftxt "$iftxt [if_operator_text $iftype]"
				if {[expr $iftype & 8] == 0} {
					set iftxt "$iftxt [::disasm::get_register $hex 5]"
				} else {
					set iftxt "$iftxt [get_word $hex 4 true]"
				}
				set iftxt "$iftxt ) then \{"
			}
			# Link part
			set linktxt " [link $hex 2][::disasm::get_hbutton_val $hex]"

			# Set part
			set direct [expr $cmdtype & 1]
			if $direct {
				set op2 [get_word $hex 2 true]
			} else {
				set op2 [::disasm::get_byte $hex 3]
			}
			set optxt [get_set_operation_text $sharedreg [::disasm::get_nibble $hex 0 1] $op2 $direct]
			# concat all parts
			set txt "(SetCLnk) $optxt ;$iftxt$linktxt"
			if {$iftxt != ""} {set txt "$txt \}"}
		}
		10 -
		11 -
		12 -
		13 {
			# CSetCLnk and CSetLnk
			# CSetLnk performs the link unconditionally
			set iftype [::disasm::get_nibble $hex 1 0]
			if {[expr $iftype & 7] == 0} {
				set iftxt ""
			} else {
				if [expr $cmdtype & 1] {
					# direct set => indirect if only
					set iftxt "if ( [::disasm::get_register $hex 4]"
					if {[expr $iftype & 8] == 0} {
						set iftxt "$iftxt [if_operator_text $iftype]"
						set iftxt "$iftxt [::disasm::get_register $hex 5]"
					} else {
						set iftxt "$iftxt <ILLEGAL IF OPERAND> <VALUE NOT ALLOWED>"
					}
					set iftxt "$iftxt ) then \{ "
				} else {
					# indirect set => direct or indirect if
					set iftxt "if ( [::disasm::get_register $hex 3]"
					set iftxt "$iftxt [if_operator_text $iftype]"
					if {[expr $iftype & 8] == 0} {
						set iftxt "$iftxt [::disasm::get_register $hex 5]"
					} else {
						set iftxt "$iftxt [get_word $hex 4 true]"
					}
					set iftxt "$iftxt ) then \{ "
				}
			}
			# Link part
			set l [link $hex 2]
			if {$l == "" } {
				set linktxt ""
			} else {
				set linktxt " ; $l[::disasm::get_hbutton_val $hex]"
			}
			# Set part
			set direct [expr $cmdtype & 1]
			if $direct {
				set op2 [get_word $hex 2 true]
			} else {
				set op2 [::disasm::get_byte $hex 2]
			}
			set optxt [get_set_operation_text [::disasm::get_nibble $hex 1 1] [::disasm::get_nibble $hex 0 1] $op2 $direct]
			# concat all parts
			if {$cmdtype >= 12} {
				set txt "$iftxt\(CSetLnk\) $optxt"
				if {$iftxt != ""} {
					set txt "$txt \}"
				}
				set txt "$txt$linktxt"
			}
			if {$cmdtype <= 11} {
				set txt "$iftxt\(CSetCLnk\) $optxt"
				set txt "$txt$linktxt"
				if {$iftxt != ""} {
					set txt "$txt \}"
				}
			}
		}
		default {
			return "ILLEGAL COMMAND \[>[format %X $cmdtype]<0 00 00 00 00 00 00 00\]"
		}
	}
	return $txt
}

#############################################################################
## Parse DVD hex command, and return the current command name, or ??? if illegal

proc cmdname {hex} {
	set cmdtype [::disasm::get_nibble $hex 0 0]
	switch $cmdtype {
		0 {
			# special/flow control commands
			switch [::disasm::get_nibble $hex 1 1] {
				0 {return "NOP"}
				1 {return "GoTo"}
				2 {return "Break"}
				3 {return "SetTmpPML"}
				default {return "???"}
			}
		}
		2 {
			# Link commands
			switch [::disasm::get_nibble $hex 1 1] {
				1 {
					if {[catch {set txt [lindex $::disasm::link1list [::disasm::get_byte $hex 7]]}] != 0} {
						return "???"
					} else {
						if {$txt == "<ILLEGAL LINK COMMAND>"} {
							return "???"
						} else {
							return $txt
						}
					}
				}
				4 {return "LinkPGCN"}
				5 {return "LinkPTT"}
				6 {return "LinkPGN"}
				7 {return "LinkCN"}
				default {return "???"}
			}
		}
		3 {
			# jump/call commands
			switch [::disasm::get_nibble $hex 1 1] {
				1 {return "Exit"}
				2 {return "JumpTT"}
				3 {return "JumpVTS_TT"}
				5 {return "JumpVTS_PTT"}
				6 {
					# JumpSS*
					switch [::disasm::get_nibble $hex 5 0] {
						0 {return "JumpSS-FP"}
						4 {return "JumpSS-VMGM-menu"}
						8 {return "JumpSS-VTSM"}
						12 {return "JumpSS-VMGM-pgc"}
						default {return "???"}
					}
				}
				8 {
					# CallSS *
					switch [::disasm::get_nibble $hex 5 0] {
						0 {return "CallSS-FP"}
						4 {return "CallSS-VMGM-menu"}
						8 {return "CallSS-VTSM"}
						12 {return "CallSS-VMGM-pgc"}
						default {return "???"}
					}
				}
				default {
					return "???"
				}
			}
		}
		4 -
		5 {
			# Set SPRM commands
			switch [::disasm::get_nibble $hex 0 1] {
				1 {return "SetSTN"}
				2 {return "SetNVTMR"}
				3 {return "SetGPRMMD"}
				4 {return "SetAMXMD"}
				6 {return "SetHL_BTN"}
				default {return "???"}
			}
		}
		6 -
		7 {return "Set"}
		8 -
		9 {return "SetCLnk"}
		10 -
		11 {return "CSetCLnk"}
		12 -
		13 {return "CSetLnk"}
		default {return "???"}
	}
	return "???"
}

#############################################################################
## Procedure:  dump_all_commands
## Dump all commands to text file
## args: header: first line of text.
proc dump_all_commands {header} {
	show_msg "Dumping VMG..."

	set out "$header"
	set out "$out\nDVD-TEXT General Name: \"[::utils::get_General_Name]\""
	set out "$out\nProvider ID: \"[::utils::get_provider_ID]\""
	set out "$out\nNumber of VTS: $::pgcs(numvts)"

	# VIDEO_TS.IFO
	set out "$out\n\n\nVMG (VIDEO_TS.IFO)"
	set out "$out\n------------------"

	set numcmds [llength $::pgcs(0,0)]
	set out "$out\n\n  VMG / FP-PGC"
	for {set c 0} {$c < $numcmds} {incr c} {
		set out "$out\n        [formatCommand [lindex $::pgcs(0,0) $c] $c]"
	}

	if {$::menus(0,numlus)==0} {
		set out "$out\n\n  VMGM: no menu."
	} else {
		set out "$out\n\n"
		for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
			set out "$out\n  VMGM / LU $lu ($::menus(0,$lu,lang)):"
			for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
				set out "$out\n\n    [::main::pgc_label_s menu 0 $lu $pgc]"
				set numcmds [llength $::menus(0,$lu,$pgc)]
				for {set c 0} {$c < $numcmds} {incr c} {
					set out "$out\n        [formatCommand [lindex $::menus(0,$lu,$pgc) $c] $c]"
				}
			}
		}
	}

	# VTS_xx_0.IFO
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		show_msg "Dumping VTS $vts..."
		set out "$out\n\n\nVTS $vts (VTS_[format %02i $vts]_0.IFO)"
		set out "$out\n--------------------"

		if {$::menus($vts,numlus)==0} {
			set out "$out\n\n  VTSM $vts: no menu."
		} else {
			set out "$out\n"
			for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
				set out "$out\n  VTSM $vts / LU $lu ($::menus($vts,$lu,lang)):"
				for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
					set out "$out\n\n    [::main::pgc_label_s menu $vts $lu $pgc]"
					set numcmds [llength $::menus($vts,$lu,$pgc)]
					for {set c 0} {$c < $numcmds} {incr c} {
						set out "$out\n        [formatCommand [lindex $::menus($vts,$lu,$pgc) $c] $c]"
					}
				}
			}
			set out "$out\n"
		}

		for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
			set out "$out\n\n  [::main::pgc_label_s pgc $vts $pgc]"
			set numcmds [llength $::pgcs($vts,$pgc)]
			for {set c 0} {$c < $numcmds} {incr c} {
				set out "$out\n        [formatCommand [lindex $::pgcs($vts,$pgc) $c] $c]"
			}
		}
	}
	set out "$out\n"
	show_msg ""
	return $out
}

#############################################################################
## Procedure:  formatCommand
## return a formatted command
## NOTE: the linenum is incremented by 1 before formatting.
proc {formatCommand} {hex linenum} {
	# format string for commands
	if {[lindex $hex 0] == "*"} {
		set type [lindex $hex 1]
		return "[string repeat "*" 10] ${type} commands:"
	} else {
		return [format {[%s] %3d  %s } [string toupper $hex] $linenum [disasm $hex]]
	}
}

}

