#############################################################################
# pgceditor.tcl Copyright 2004 r0lZ
# The PGC Header editor stuff.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::pgced {

namespace export edit_pgc


proc BCDtime {bcd showfps} {
	set pbtimetxt           [format  %i [expr  [get_int bcd 0 1] / 16]]
	set pbtimetxt $pbtimetxt[format  %i [expr  [get_int bcd 0 1] & 15]]
	set pbtimetxt $pbtimetxt:[format %i [expr  [get_int bcd 1 1] / 16]]
	set pbtimetxt $pbtimetxt[format  %i [expr  [get_int bcd 1 1] & 15]]
	set pbtimetxt $pbtimetxt:[format %i [expr  [get_int bcd 2 1] / 16]]
	set pbtimetxt $pbtimetxt[format  %i [expr  [get_int bcd 2 1] & 15]]
	set pbtimetxt $pbtimetxt.[format %i [expr ([get_int bcd 3 1]& 0X30)/16]]
	set pbtimetxt $pbtimetxt[format  %i [expr  [get_int bcd 3 1] & 15]]
	if $showfps {
		set fpsmode [expr [get_int bcd 3 1] & 0XC0]
		switch $fpsmode {
			192 {set pbtimetxt "$pbtimetxt (at 30 fps)"}
			64  {set pbtimetxt "$pbtimetxt (at 25 fps)"}
			default {set pbtimetxt "$pbtimetxt (illegal FPS)"}
		}
	}
	return $pbtimetxt
}

proc AddBCDtime {bcd1 bcd2} {
	set fpsmode [expr [get_int bcd1 3 1] & 0XC0]
	switch $fpsmode {
		192 {set fps 30}
		64  {set fps 25}
		default {set fps 0}
	}
	if {$fps == 0} {return "\000\000\000\000"}

	set f1 [expr ((([get_int bcd1 3 1]& 0X30)/16)*10) + ([get_int bcd1 3 1] & 15)]
	set s1 [expr  (([get_int bcd1 2 1] / 16) * 10) + ([get_int bcd1 2 1] & 15)]
	set m1 [expr  (([get_int bcd1 1 1] / 16) * 10) + ([get_int bcd1 1 1] & 15)]
	set h1 [expr  (([get_int bcd1 0 1] / 16) * 10) + ([get_int bcd1 0 1] & 15)]

	set f2 [expr ((([get_int bcd2 3 1]& 0X30)/16)*10) + ([get_int bcd2 3 1] & 15)]
	set s2 [expr  (([get_int bcd2 2 1] / 16) * 10) + ([get_int bcd2 2 1] & 15)]
	set m2 [expr  (([get_int bcd2 1 1] / 16) * 10) + ([get_int bcd2 1 1] & 15)]
	set h2 [expr  (([get_int bcd2 0 1] / 16) * 10) + ([get_int bcd2 0 1] & 15)]

	set f [expr $f1 + $f2]
	set s [expr $s1 + $s2]
	set m [expr $m1 + $m2]
	set h [expr $h1 + $h2]

	while {$f >= $fps} {
		set f [expr $f - $fps]
		incr s
	}
	while {$s >= 60} {
		set s [expr $s - 60]
		incr m
	}
	while {$m >= 60} {
		set m [expr $m - 60]
		incr h
	}

	set f "[expr ((($f / 10) * 16) + ($f % 10)) | $fpsmode]"
	set s "[expr (($s / 10) * 16) + ($s % 10)]"
	set m "[expr (($m / 10) * 16) + ($m % 10)]"
	set h "[expr (($h / 10) * 16) + ($h % 10)]"

	return "[format %c $h][format %c $m][format %c $s][format %c $f]"
}

#### Main entry for PgcEditor
proc edit_pgc {pgcheader numprograms programmap numcells cellpbtable cellpostable PTTlist} {
	set ::pgced::__pgcheader    $pgcheader
	set ::pgced::__programmap   $programmap
	set ::pgced::__cellpbtable  $cellpbtable
	set ::pgced::__cellpostable $cellpostable
	set ::pgced::__PTTlist      $PTTlist

	# get editables values from the header's binary data

	set pgcpbtime [string range $::pgced::__pgcheader 4 7]
	set ::pgced::pbtimetxt [BCDtime $pgcpbtime true]

	set ::pgced::pbmode    [get_int ::pgced::__pgcheader 0XA2 1]
	if {$::pgced::pbmode == 0} {
		set pbmodetxt "sequential"
	} else {
		set pbprogs [expr $::pgced::pbmode & 0X7F]
		if {[expr $::pgced::pbmode & 0X80] == 0} {
			set pbmodetxt "random: $pbprogs programs"
		} else {
			set pbmodetxt "shuffle: $pbprogs programs"
		}
	}
	set ::pgced::pbmodetxt $pbmodetxt

	set ::pgced::puos      [get_int ::pgced::__pgcheader 8 4]
	set ::pgced::nextpgcn  [get_int ::pgced::__pgcheader 0X9C 2]
	set ::pgced::prevpgcn  [get_int ::pgced::__pgcheader 0X9E 2]
	set ::pgced::gouppgcn  [get_int ::pgced::__pgcheader 0XA0 2]
	set ::pgced::stilltime [get_int ::pgced::__pgcheader 0XA3 1]

	# build interface
	set w ".pgceditor"
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm iconname $w Dialog
	wm protocol $w WM_DELETE_WINDOW {set ::pgced::done 0}
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	set stdcursor [$w cget -cursor]
	bind $w <Escape> {set ::pgced::done 0}

	frame $w.f1
	pack  $w.f1 -pady 2 -anchor w

	# misc PGC values
	frame $w.f1.fl
	pack  $w.f1.fl -fill x -padx 2 -pady 2 -side left
	frame $w.f1.fl.mf
	pack  $w.f1.fl.mf -padx 2
	label $w.f1.fl.mf.l0 -text "Playback Time "
	entry $w.f1.fl.mf.e0 -textvariable ::pgced::pbtimetxt -state readonly -justify center
	label $w.f1.fl.mf.l1 -text "Playback Mode "
	entry $w.f1.fl.mf.e1 -textvariable ::pgced::pbmodetxt -state readonly -justify center
	label $w.f1.fl.mf.l2 -text "Prohibited User Operations "
	entry $w.f1.fl.mf.e2 -textvariable ::pgced::puos -width 10 -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0x1FFFFFF {%d} {%P}"
	button $w.f1.fl.mf.b2 -text "..." -pady 0 \
			-command {::pgced::puos_editor "Prohibited User Operations" .pgceditor}
	bind   $w.f1.fl.mf.b2 <3> {set ::pgced::puos 0}
	label $w.f1.fl.mf.l3 -text "Previous PGCN "
	entry $w.f1.fl.mf.e3 -textvariable ::pgced::prevpgcn  -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0XFFFF {%d} {%P}"
	label $w.f1.fl.mf.l4 -text "Next PGCN "
	entry $w.f1.fl.mf.e4 -textvariable ::pgced::nextpgcn  -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0XFF {%d} {%P}"
	label $w.f1.fl.mf.l5 -text "Go Up (Return) PGCN "
	entry $w.f1.fl.mf.e5 -textvariable ::pgced::gouppgcn  -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0XFF {%d} {%P}"
	label $w.f1.fl.mf.l6 -text "PGC Still Time (255=infinite) "
	entry $w.f1.fl.mf.e6 -textvariable ::pgced::stilltime  -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0XFF {%d} {%P}"
	grid  $w.f1.fl.mf.l0 -row 0 -column 0 -sticky e -ipady 8
	grid  $w.f1.fl.mf.e0 -row 0 -column 1 -sticky ew -columnspan 2
	grid  $w.f1.fl.mf.l1 -row 1 -column 0 -sticky e -ipady 2
	grid  $w.f1.fl.mf.e1 -row 1 -column 1 -sticky ew -columnspan 2
	grid  $w.f1.fl.mf.l2 -row 2 -column 0 -sticky e -ipady 8
	grid  $w.f1.fl.mf.e2 -row 2 -column 1 -sticky ew
	grid  $w.f1.fl.mf.b2 -row 2 -column 2 -sticky ew
	grid  $w.f1.fl.mf.l3 -row 3 -column 0 -sticky e -ipady 2
	grid  $w.f1.fl.mf.e3 -row 3 -column 1 -sticky ew -columnspan 2
	grid  $w.f1.fl.mf.l4 -row 4 -column 0 -sticky e -ipady 2
	grid  $w.f1.fl.mf.e4 -row 4 -column 1 -sticky ew -columnspan 2
	grid  $w.f1.fl.mf.l5 -row 5 -column 0 -sticky e -ipady 2
	grid  $w.f1.fl.mf.e5 -row 5 -column 1 -sticky ew -columnspan 2
	grid  $w.f1.fl.mf.l6 -row 6 -column 0 -sticky e -ipady 8
	grid  $w.f1.fl.mf.e6 -row 6 -column 1 -sticky ew -columnspan 2

	# audio streams control
	frame $w.f1.mfa -borderwidth 2 -relief groove
	pack  $w.f1.mfa -padx 2 -ipadx 4 -ipady 2 -side left
	label $w.f1.mfa.tit -text "Audio\nStreams" -pady 0
	grid  $w.f1.mfa.tit -row 0 -column 0 -columnspan 2
	for {set a 1} {$a <= 8} {incr a} {
		label $w.f1.mfa.al$a -text "$a "
		entry $w.f1.mfa.ab$a -textvariable ::pgced::audiostream$a -state disabled \
				-disabledforeground black -relief raised -width 5 \
				-font $::config(fixed_width_font) -cursor $stdcursor
		bind  $w.f1.mfa.ab$a <1> "::pgced::set_audio ::pgced::audiostream$a"
		grid  $w.f1.mfa.al$a -row $a -column 0 -sticky e
		grid  $w.f1.mfa.ab$a -row $a -column 1 -sticky ew
	}

	# sub-picture streams control
	frame $w.f1.mfs -borderwidth 2 -relief groove
	pack  $w.f1.mfs -padx 2 -pady 1 -ipadx 4 -ipady 2 -side left
	label $w.f1.mfs.tit -text "Sub-picture Streams\n(for 4:3, wide, letterbox, pan&scan)" -pady 0
	grid  $w.f1.mfs.tit -row 0 -column 0 -columnspan 8
	for {set s 1} {$s <= 32} {incr s} {
		label $w.f1.mfs.al$s -text "$s "
		entry $w.f1.mfs.ab$s -textvariable ::pgced::subpicstream$s -state disabled \
				-disabledforeground black -relief raised -width 11 \
				-font $::config(fixed_width_font) -cursor $stdcursor
		bind  $w.f1.mfs.ab$s <1> "::pgced::set_subpic ::pgced::subpicstream$s"
		set row [expr (($s-1) % 8) + 1]
		set col [expr (($s-1) / 8) * 2]
		grid  $w.f1.mfs.al$s -row $row -column $col -sticky e
		grid  $w.f1.mfs.ab$s -row $row -column [expr $col+1] -sticky ew
	}

	# bottom frame
	frame  $w.f2
	pack   $w.f2 -expand 1 -fill both

	# color table
	frame $w.f2.mfc -borderwidth 2 -relief groove
	pack  $w.f2.mfc -side left -fill x -ipadx 2 -ipady 2 -padx 4
	label $w.f2.mfc.l0 -text "Sub-picture\nColor Lookup Table "
	pack  $w.f2.mfc.l0 -fill x
	frame $w.f2.mfc.f1
	pack  $w.f2.mfc.f1
	set col 0
	set row 0
	for {set c 0} {$c < 16} {incr c} {
		button $w.f2.mfc.f1.b$c -width 2 -pady 0 -command "::pgced::set_color $c $w.f2.mfc.f1.b$c"
		grid   $w.f2.mfc.f1.b$c -row $row -column $col -ipadx 3 -ipady 3 -padx 1 -pady 1
		incr col
		if {$col == 4} {
			set col 0
			incr row
		}
	}
	frame $w.f2.mfc.pad -height 4
	pack  $w.f2.mfc.pad
	frame $w.f2.mfc.f2
	pack  $w.f2.mfc.f2
	button $w.f2.mfc.f2.bc -text "copy all"  -width 8 -command ::pgced::copy_colors  -pady 0
	button $w.f2.mfc.f2.bp -text "paste all" -width 8 -command ::pgced::paste_colors -pady 0
	pack   $w.f2.mfc.f2.bc -side left -padx 3
	pack   $w.f2.mfc.f2.bp -side left -padx 2


	### Cells table
	frame  $w.f2.fc -borderwidth 2 -relief groove
	pack   $w.f2.fc -side left -padx 2 -expand 1 -fill both -anchor w
	canvas $w.f2.fc.c -yscrollcommand "$w.f2.fc.vs set" -height 8 \
			-highlightcolor [$w.f2.fc cget -background]
	pack   $w.f2.fc.c -side left -expand 1 -fill both
	scrollbar $w.f2.fc.vs -command "$w.f2.fc.c yview" -orient vert
	pack   $w.f2.fc.vs -side left -fill y
	frame  $w.f2.fc.c.f	;# destroyed later

	### buttons
	frame  $w.fb
	pack   $w.fb -anchor e -padx 4 -pady 4 -fill x
	button $w.fb.ss -text " Current domain streams attributes " \
			-command "::streams::set_streams_attribs $w"
	pack   $w.fb.ss -side left -padx 2
	frame  $w.fb.pad
	pack   $w.fb.pad -side left -fill x -expand 1
	button $w.fb.ok -text "OK" -width 10 -command {set ::pgced::done 1}
	pack   $w.fb.ok -side left -padx 3
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::pgced::done 0}
	pack   $w.fb.cancel -side left -padx 2

	wm title $w "PGC Editor - [$::widget(main_cmd_lab) cget -text]"

	for {set c 0} {$c < 16} {incr c} {
		set yc_color [string range $::pgced::__pgcheader [expr 0XA4 + ($c*4)] [expr 0XA4 + ($c*4) + 3]]
		$w.f2.mfc.f1.b$c configure -background [yc2rgb $yc_color]
	}
	for {set a 1} {$a <= 8} {incr a} {
		binary scan [string range $::pgced::__pgcheader [expr 0X0A + ($a*2)] [expr 0X0A + ($a*2)]] c audio_status
		if {$audio_status == 0} {
			set ::pgced::audiostream$a "none"
		} else {
			set ::pgced::audiostream$a [expr $audio_status & 7]
		}
	}
	for {set s 1} {$s <= 32} {incr s} {
		binary scan [string range $::pgced::__pgcheader [expr 0X18 + ($s*4)] [expr 0X18 + ($s*4) + 3]] cccc s1 s2 s3 s4
		if {$s1 == 0} {
			set ::pgced::subpicstream$s "none"
		} else {
			set ::pgced::subpicstream$s [format "%2d %2d %2d %2d" [expr $s1 & 0X1F] $s2 $s3 $s4]
		}
	}


	set end_time [build_cell_table $w $numprograms $numcells]
 
	# window geometry
	#set maxheight [winfo height $w.f2.fc.c.f]
	#if {$maxheight > 200} {set maxheight 200}
	#$w.f2.fc.c configure -height $maxheight
	update
	set ww [winfo reqwidth $w]
	set wh [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-$ww)/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-$wh)/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	$w.f2.fc.c configure -height 1
	wm geometry $w ${ww}x$wh+$wx+$wy
	wm resizable $w 1 1

	if {! [winfo ismapped $w]} {
		wm deiconify $w
	}
	grab set $w

	if {$end_time != -1 && $end_time != $pgcpbtime} {
		set txt "WARNING: The PGC Playback Time ([BCDtime $pgcpbtime true]) doesn't match"
		set txt "$txt\nthe total of the cells times ([BCDtime $end_time true])."
		set txt "$txt\n\nFix it?"
		set r [tk_messageBox -title "Pgc Editor" -icon warning \
			-message $txt -parent . -type yesno -default yes]
		if {$r == "yes"} {
			set ::pgced::__pgcheader "[string range $::pgced::__pgcheader 0 3]$end_time[string range $::pgced::__pgcheader 8 end]"
			set ::pgced::pbtimetxt [BCDtime $end_time true]
		}
	}

	set ::pgced::done -1
	tkwait variable ::pgced::done
	grab release $w
	wm withdraw $w

	if $::pgced::done {
		accept_modifs $w $numcells
		return [list $::pgced::__pgcheader $::pgced::__cellpbtable $::pgced::__cellpostable]
	} else {
		return ""
	}
}

proc accept_modifs {w numcells} {
	if {$::pgced::puos      != ""} {set_int ::pgced::__pgcheader 8    [expr $::pgced::puos] 4}
	if {$::pgced::nextpgcn  != ""} {set_int ::pgced::__pgcheader 0X9C [expr $::pgced::nextpgcn] 2}
	if {$::pgced::prevpgcn  != ""} {set_int ::pgced::__pgcheader 0X9E [expr $::pgced::prevpgcn] 2}
	if {$::pgced::gouppgcn  != ""} {set_int ::pgced::__pgcheader 0XA0 [expr $::pgced::gouppgcn] 2}
	if {$::pgced::stilltime != ""} {set_int ::pgced::__pgcheader 0XA3 [expr $::pgced::stilltime] 1}
	for {set c 0} {$c < 16} {incr c} {
		set rgb_color [$w.f2.mfc.f1.b$c cget -background]
		set_int ::pgced::__pgcheader [expr 0XA4 + (4*$c)] [rgb2yc $rgb_color] 4
	}
	for {set a 1} {$a <= 8} {incr a} {
		set varname ::pgced::audiostream$a
		if {[subst $$varname] == "none"} {
			set as 0
		} else {
			set as [expr [subst $$varname] | 0X80]
		}
		set_int ::pgced::__pgcheader [expr 0X0A + ($a*2)] $as 1
	}
	for {set s 1} {$s <= 32} {incr s} {
		set varname ::pgced::subpicstream$s
		if {[subst $$varname] == "none"} {
			set ss 0
		} else {
			set sl [subst $$varname]
			set ss [expr 0X80000000 + \
					([lindex $sl 0] * 0X1000000) + \
					([lindex $sl 1] * 0X10000) + \
					([lindex $sl 2] * 0X100) + \
					[lindex $sl 3]]
		}
		set_int ::pgced::__pgcheader [expr 0X18 + ($s*4)] $ss 4
	}

	for {set c 1} {$c <= $numcells} {incr c} {
		set cellidx1 [expr ($c - 1) * 24]
		set_int ::pgced::__cellpbtable $cellidx1 [$w.f2.fc.c.f.w_${c}_3 cget -text] 1
		set_int ::pgced::__cellpbtable [expr $cellidx1 + 1] [subst \$::pgced::var_${c}_5] 1
		set_int ::pgced::__cellpbtable [expr $cellidx1 + 2] [subst \$::pgced::var_${c}_6] 1
		set_int ::pgced::__cellpbtable [expr $cellidx1 + 3] [subst \$::pgced::var_${c}_7] 1
	}
}

### build Cells table (+ PTT and Programs infos)
proc build_cell_table {w numprograms numcells} {

	destroy $w.f2.fc.c.f
	frame $w.f2.fc.c.f
	pack  $w.f2.fc.c.f -anchor nw
	$w.f2.fc.c create window 0 0 -window $w.f2.fc.c.f -anchor nw

	if {$numcells == 0} {
		label $w.f2.fc.c.f.w_0_0 -text "No Cells"
		pack $w.f2.fc.c.f.w_0_0
		set end_time -1
	} else {
		label $w.f2.fc.c.f.w_0_0 -text "Chap.\n(PTT)"
		label $w.f2.fc.c.f.w_0_1 -text "Prog."
		label $w.f2.fc.c.f.w_0_2 -text "Cell"
		label $w.f2.fc.c.f.w_0_3 -text "Type\nFlags"
		label $w.f2.fc.c.f.w_0_4 -text "Layer\nBreak"
		label $w.f2.fc.c.f.w_0_5 -text "Res-\ntricted"
		label $w.f2.fc.c.f.w_0_6 -text "Still\nTime"
		label $w.f2.fc.c.f.w_0_7 -text "Cell\nCmd"
		label $w.f2.fc.c.f.w_0_8 -text "Playback\nTime"
		label $w.f2.fc.c.f.w_0_9 -text "End\nTime"
		label $w.f2.fc.c.f.w_0_10 -text "Entry\nVOBU\nsector"
		label $w.f2.fc.c.f.w_0_11 -text "First\nILVU\nEnd"
		label $w.f2.fc.c.f.w_0_12 -text "Last\nVOBU\nStart"
		label $w.f2.fc.c.f.w_0_13 -text "Last\nVOBU\nEnd"
		label $w.f2.fc.c.f.w_0_14 -text "VOB/Cell\nID"

		for {set c 1} {$c <= $numcells} {incr c} {
			set cell_program($c) ""
			set prog_chapter($c) ""
		}
		for {set p 1} {$p <= $numprograms} {incr p} {
			set cell_program([get_int ::pgced::__programmap [expr $p-1] 1]) $p
		}
		for {set c 1} {$c <= $numcells} {incr c} {
			for {set ptt 0} {$ptt < [llength $::pgced::__PTTlist]} {incr ptt} {
				if {[lindex $::pgced::__PTTlist $ptt] == $cell_program($c)} {
					set prog_chapter($c) [expr $ptt+1]
				}
			}
		}

		set end_time "\000\000\000\000"
		for {set c 1} {$c <= $numcells} {incr c} {
			set cellidx1 [expr ($c - 1) * 24]
			set cellidx2 [expr ($c - 1) * 4]

			label $w.f2.fc.c.f.w_${c}_0 -text $prog_chapter($c)
			label $w.f2.fc.c.f.w_${c}_1 -text $cell_program($c)
			label $w.f2.fc.c.f.w_${c}_2 -text $c

			set type [get_int ::pgced::__cellpbtable $cellidx1 1]
			button $w.f2.fc.c.f.w_${c}_3 -text $type -width 3 \
					-anchor e -pady 0 \
					-command "::pgced::cell_type_editor $w.f2.fc.c.f.w_${c}_3 ::pgced::var_${c}_4"
			if {[expr $type & 8] == 8} {
				set [subst ::pgced::var_${c}_4] 0
			} else {
				set [subst ::pgced::var_${c}_4] 1
			}
			checkbutton $w.f2.fc.c.f.w_${c}_4 -variable ::pgced::var_${c}_4 \
					-command "::pgced::set_layer_break ::pgced::var_${c}_4 $w.f2.fc.c.f.w_${c}_3"
			if {$c == 1} {$w.f2.fc.c.f.w_${c}_4 configure -state disabled}
			set  ::pgced::var_${c}_5 [expr [get_int ::pgced::__cellpbtable [expr $cellidx1 + 1] 1] & 64]
			checkbutton $w.f2.fc.c.f.w_${c}_5 -variable ::pgced::var_${c}_5 \
					-variable ::pgced::var_${c}_5 -onvalue 64 -offvalue 0
			set  ::pgced::var_${c}_6 [get_int ::pgced::__cellpbtable [expr $cellidx1 + 2] 1]
			entry $w.f2.fc.c.f.w_${c}_6 -textvariable ::pgced::var_${c}_6 -width 3 -justify right
			set  ::pgced::var_${c}_7 [get_int ::pgced::__cellpbtable [expr $cellidx1 + 3] 1]
			entry $w.f2.fc.c.f.w_${c}_7 -textvariable ::pgced::var_${c}_7 -width 3 -justify right
			set cell_time [string range $::pgced::__cellpbtable [expr $cellidx1 + 4] [expr $cellidx1 + 7]]
			label $w.f2.fc.c.f.w_${c}_8 -text [BCDtime $cell_time false]
			set end_time [AddBCDtime $cell_time $end_time]
			label $w.f2.fc.c.f.w_${c}_9 -text [BCDtime $end_time false]
			label $w.f2.fc.c.f.w_${c}_10 -text [get_int ::pgced::__cellpbtable [expr $cellidx1 +  8] 4]
			label $w.f2.fc.c.f.w_${c}_11 -text [get_int ::pgced::__cellpbtable [expr $cellidx1 + 12] 4]
			label $w.f2.fc.c.f.w_${c}_12 -text [get_int ::pgced::__cellpbtable [expr $cellidx1 + 16] 4]
			label $w.f2.fc.c.f.w_${c}_13 -text [get_int ::pgced::__cellpbtable [expr $cellidx1 + 20] 4]
			set vobid  [get_int ::pgced::__cellpostable $cellidx2 2]
			set cellid [get_int ::pgced::__cellpostable [expr $cellidx2 + 3] 1]
			button $w.f2.fc.c.f.w_${c}_14 -pady 0 \
					-text "$vobid / $cellid" \
					-command "::pgced::accept_modifs $w $numcells ; \
							::pgced::reassign_VOB_ID $w $c $vobid $cellid $numprograms $numcells"
		}
		for {set c 0} {$c <= $numcells} {incr c} {
			for {set f 0} {$f <= 14} {incr f} {
				grid $w.f2.fc.c.f.w_${c}_$f -row $c -column $f
			}
			grid configure $w.f2.fc.c.f.w_${c}_14 -sticky ew -padx 2
		}
		#for {set f 0} {$f <= 14} {incr f} {
		#	grid columnconf $w.f2.fc.c.f $f -minsize 28 -pad 2
		#}
	}

	# inform scrollbar of the height of the canvas's internal window
	update
	$w.f2.fc.c configure -scrollregion "0 0 0 [winfo height $w.f2.fc.c.f]" \
			-width [winfo width $w.f2.fc.c.f]
	
	return $end_time
}

proc value_entry_validate {min max action val} {
	if {$action != 0} {
		if {$val=="0"||$val=="0X"||$val=="0X"} {
			if {$min>0 && $val=="0"} bell
			return true
		}
		if {![string is integer $val]} {
			bell
			return false
		}
		if {$val < $min} {
			bell
			return false
		}
		if {$val > $max} {
			bell
			return false
		}
	}
	return true
}

proc rgb2yc {rgb_color} {
	if {[string range $rgb_color 0 0] == "#"} {
		set rgb_color [string range $rgb_color 1 end]
	}
	set r [expr 0X[string range $rgb_color 0 1]]
	set g [expr 0X[string range $rgb_color 2 3]]
	set b [expr 0X[string range $rgb_color 4 5]]

	set y  [expr round(( 0.257*$r) + (0.504*$g) + (0.098*$b) + 16)]
	set cr [expr round(( 0.439*$r) - (0.368*$g) - (0.071*$b) + 128)]
	set cb [expr round((-0.148*$r) - (0.291*$g) + (0.439*$b) + 128)]

	if {$y  < 0} {set y  0}
	if {$cr < 0} {set cr 0}
	if {$cb < 0} {set cb 0}
	if {$y  > 255} {set y  255}
	if {$cr > 255} {set cr 255}
	if {$cb > 255} {set cb 255}

	return [expr ($y*0X10000) + ($cr*0X100) + $cb]
}

proc yc2rgb {yc_color} {
	if {[string length $yc_color] == 4} {
		set yc_color [string range $yc_color 1 end]
	}
	binary scan $yc_color ccc y cr cb
	set y  [expr ($y  + 0X100) % 0X100]
	set cr [expr ($cr + 0X100) % 0X100]
	set cb [expr ($cb + 0X100) % 0X100]

	set r [expr round((1.164*($y-16)) + (1.596*($cr-128)))]
	set g [expr round((1.164*($y-16)) - (0.813*($cr-128)) - (0.392*($cb-128)))]
	set b [expr round((1.164*($y-16)) + (2.017*($cb-128)))]

	if {$r < 0} {set r 0}
	if {$g < 0} {set g 0}
	if {$b < 0} {set b 0}
	if {$r > 255} {set r 255}
	if {$g > 255} {set g 255}
	if {$b > 255} {set b 255}

	return [format "#%02X%02X%02X" $r $g $b]
}

proc set_color {colornum w} {
	set newcolor [tk_chooseColor -initialcolor [$w cget -background] -parent .pgceditor -title "Select color #$colornum"]
	if {$newcolor == ""} return
	$w configure -background $newcolor
}

proc copy_colors {} {
	set colors {}
	for {set c 0} {$c < 16} {incr c} {
		lappend colors [.pgceditor.f2.mfc.f1.b$c cget -background]
	}
	clipboard clear
	clipboard append $colors
}

proc paste_colors {} {
	set err [catch {set colors [clipboard get]}]
	if {$err || [llength $colors] != 16} {
		bell
		return
	}
	for {set c 0} {$c < 16} {incr c} {
		.pgceditor.f2.mfc.f1.b$c configure -background [lindex $colors $c]
	}
}

set puos_labels {
	{Time play or search}
	{PTT play or search}
	{Title play}
	{Stop}
	{GoUp  (Not used in PGCs) }
	{Time or PTT search}
	{TopPG or PrevPG search}
	{NextPG search}
	{Forward scan}
	{Backward scan}
	{Menu call - Title}
	{Menu call - Root}
	{Menu call - Subpicture}
	{Menu call - Audio}
	{Menu call - Angle}
	{Menu call - PTT}
	{Resume}
	{Button select or activate}
	{Still off}
	{Pause on}
	{Audio stream change}
	{Subpicture stream change}
	{Angle change}
	{Karaoke audio mix change}
	{Video presentation mode change}
}

proc puos_editor {label parent} {
	# puos labels
	# note: GoUp (bit 4) is not used in PGC

	set w ".puoseditor"
	if {![winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm iconname $w Dialog
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::puosdone 0}
		bind $w <Escape> "set ::pgced::puosdone 0"
		wm transient $w $parent
		if {$::tcl_platform(platform) == "windows"} {
			catch {wm attributes $w -toolwindow 1}
		}
		wm title $w "Prohibited User Operations"

		label $w.lab -text $label
		pack  $w.lab -fill x -padx 4

		frame $w.f1
		pack  $w.f1 -fill x -padx 4

		set puon 0
		foreach p $::pgced::puos_labels {
			checkbutton $w.f1.puo$puon -text "$p ([expr round(pow(2,$puon))])" -anchor w -pady 0 -borderwidth 0 -variable ::pgced::puo_$puon
			pack $w.f1.puo$puon -fill x -pady 0
			incr puon
		}
		#$w.f1.puo4 configure -state disabled	;# GoUp flag normally not used in PGC (?)
		frame  $w.fb
		pack   $w.fb -fill x -padx 2 -pady 4
		button $w.fb.clear -text "Clear all" -command {set ::pgced::tmppuos 0; ::pgced::set_puos_vars}
		pack   $w.fb.clear -side left -padx 5
		frame  $w.fb.pad
		pack   $w.fb.pad -side left -expand 1 -fill x
		button $w.fb.ok -text "OK" -width 6 -command {set ::pgced::puosdone 1}
		pack   $w.fb.ok -side left
		button $w.fb.cancel -text "Cancel" -width 6 -command {set ::pgced::puosdone 0}
		pack   $w.fb.cancel -side left -padx 5

		update
		set wx [expr [winfo rootx $parent]+(([winfo width  $parent]-[winfo reqwidth  $w])/2)]
		set wy [expr [winfo rooty $parent]+(([winfo height $parent]-[winfo reqheight $w])/2)-25]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}
	set ::pgced::tmppuos $::pgced::puos
	#set ::pgced::tmppuos [expr $::pgced::puos & 0X1FFFFEF]	;# disable GoUp flag, not used in PGC (?)
	::pgced::set_puos_vars
	$w.lab configure -text $label

	wm deiconify $w
	grab set $w

	set ::pgced::puosdone -1
	tkwait variable ::pgced::puosdone
	grab release $w
	if {$parent != "."} {
		grab set $parent
	}
	wm withdraw $w
	if $::pgced::puosdone {
		set ::pgced::puos 0
		for {set p 0} {$p <= 24} {incr p} {
			set varname ::pgced::puo_$p
			if [subst $$varname] {
				set ::pgced::puos [expr $::pgced::puos | round(pow(2,$p))]
			}
		}
	}
	return $::pgced::puos
}

proc set_puos_vars {} {
	for {set p 0} {$p <= 24} {incr p} {
		if {[expr $::pgced::tmppuos & round(pow(2,$p))] == 0} {
			set ::pgced::puo_$p 0
		} else {
			set ::pgced::puo_$p 1
		}
	}
}

proc set_audio {streamvar} {
	set w .audiosetup
	if {! [winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::audiodone 0}
		bind $w <Escape> "set ::pgced::audiodone 0"
		wm title $w "Audio stream setup"
		wm resizable $w 0 0
		if {$::tcl_platform(platform) == "windows"} {
			catch {wm attributes $w -toolwindow 1}
		}
		label $w.l -text "Audio stream number:"
		pack  $w.l
		scale $w.s -from 0 -to 7 -variable "::pgced::audio_status"
		pack  $w.s -fill x -expand 1 -padx 4

		frame  $w.fb
		pack   $w.fb -padx 2 -pady 4 -anchor e
		button $w.fb.dis -text "Disable" -width 7 -command {set ::pgced::audiodone 2}
		pack   $w.fb.dis -side left -padx 4
		button $w.fb.ok -text "Set" -width 7 -command {set ::pgced::audiodone 1}
		pack   $w.fb.ok -side left
		button $w.fb.cancel -text "Cancel" -width 7 -command {set ::pgced::audiodone 0}
		pack   $w.fb.cancel -side left -padx 4

		update
		set wx [expr [winfo rootx .pgceditor]+(([winfo width  .pgceditor]-[winfo reqwidth  $w])/2)]
		set wy [expr [winfo rooty .pgceditor]+(([winfo height .pgceditor]-[winfo reqheight $w])/2)-25]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}
	if {[subst $$streamvar] == "none"} {
		set ::pgced::audio_status 0
	} else {
		set ::pgced::audio_status [subst $$streamvar]
	}

	wm deiconify $w
	grab set $w

	set ::pgced::audiodone -1
	tkwait variable ::pgced::audiodone
	grab release $w
	grab set .pgceditor
	wm withdraw $w
	switch $::pgced::audiodone {
		1 {set $streamvar $::pgced::audio_status}
		2 {set $streamvar "none"}
	}
	return
}

proc set_subpic {streamvar} {
	set w .subpicsetup
	if {! [winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::subpicdone 0}
		bind $w <Escape> "set ::pgced::subpicdone 0"
		wm title $w "Subpic streams setup"
		wm resizable $w 0 0
		if {$::tcl_platform(platform) == "windows"} {
			catch {wm attributes $w -toolwindow 1}
		}
		label $w.l -text "Sub-picture stream numbers:"
		pack  $w.l

		frame $w.f
		pack  $w.f -padx 4
		label $w.f.l1 -text "\n4:3"
		scale $w.f.s1 -from 0 -to 15 -variable "::pgced::subpic_1_status" -length 170
		label $w.f.l2 -text "\nwide"
		scale $w.f.s2 -from 0 -to 15 -variable "::pgced::subpic_2_status"
		label $w.f.l3 -text "\nletterbox"
		scale $w.f.s3 -from 0 -to 15 -variable "::pgced::subpic_3_status"
		label $w.f.l4 -text "\npan&scan"
		scale $w.f.s4 -from 0 -to 15 -variable "::pgced::subpic_4_status"
		grid  $w.f.l1 -row 0 -column 0 -sticky e
		grid  $w.f.s1 -row 0 -column 1 -sticky ew
		grid  $w.f.l2 -row 1 -column 0 -sticky e
		grid  $w.f.s2 -row 1 -column 1 -sticky ew
		grid  $w.f.l3 -row 2 -column 0 -sticky e
		grid  $w.f.s3 -row 2 -column 1 -sticky ew
		grid  $w.f.l4 -row 3 -column 0 -sticky e
		grid  $w.f.s4 -row 3 -column 1 -sticky ew

		frame  $w.fb
		pack   $w.fb -padx 2 -pady 4 -anchor e
		button $w.fb.dis -text "Disable" -width 7 -command {set ::pgced::subpicdone 2}
		pack   $w.fb.dis -side left -padx 4
		button $w.fb.ok -text "Set" -width 7 -command {set ::pgced::subpicdone 1}
		pack   $w.fb.ok -side left
		button $w.fb.cancel -text "Cancel" -width 7 -command {set ::pgced::subpicdone 0}
		pack   $w.fb.cancel -side left -padx 4

		update
		set wx [expr [winfo rootx .pgceditor]+(([winfo width  .pgceditor]-[winfo reqwidth  $w])/2)]
		set wy [expr [winfo rooty .pgceditor]+(([winfo height .pgceditor]-[winfo reqheight $w])/2)-25]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}
	if {[subst $$streamvar] == "none"} {
		set ::pgced::subpic_1_status 0
		set ::pgced::subpic_2_status 0
		set ::pgced::subpic_3_status 0
		set ::pgced::subpic_4_status 0
	} else {
		set sublist [subst $$streamvar]
		set ::pgced::subpic_1_status [lindex $sublist 0]
		set ::pgced::subpic_2_status [lindex $sublist 1]
		set ::pgced::subpic_3_status [lindex $sublist 2]
		set ::pgced::subpic_4_status [lindex $sublist 3]
	}
	wm deiconify $w
	grab set $w

	set ::pgced::subpicdone -1
	tkwait variable ::pgced::subpicdone
	grab release $w
	grab set .pgceditor
	wm withdraw $w
	switch $::pgced::subpicdone {
		1 {set $streamvar [format "%2d %2d %2d %2d" $::pgced::subpic_1_status \
				$::pgced::subpic_2_status $::pgced::subpic_3_status $::pgced::subpic_4_status]}
		2 {set $streamvar "none"}
	}
	return
}

proc cell_type_editor {parentbutton layerbreakvar} {

	set w ".celltypeeditor"
	if {![winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm iconname $w Dialog
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::celltypedone 0}
		bind $w <Escape> "set ::pgced::celltypedone 0"
		wm transient $w .pgceditor
		if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
		wm title $w "Cell Type Flags"

		frame $w.f1
		pack  $w.f1 -fill x -padx 4 -pady 4

		checkbutton $w.f1.c0 -text "Seamless angle linked in DSI (1)" -anchor w \
				-borderwidth 0 -variable ::pgced::celltype_0 -offvalue 0 -onvalue 1
		pack $w.f1.c0 -fill x
		checkbutton $w.f1.c1 -text "SCR discontinuity (2)" -anchor w \
				-borderwidth 0 -variable ::pgced::celltype_1 -offvalue 0 -onvalue 2
		pack $w.f1.c1 -fill x
		checkbutton $w.f1.c2 -text "Interleaved (4)" -anchor w \
				-borderwidth 0 -variable ::pgced::celltype_2 -offvalue 0 -onvalue 4
		pack $w.f1.c2 -fill x
		checkbutton $w.f1.c3 -text "Seamless playback linked in PCI (8)" -anchor w \
				-borderwidth 0 -variable ::pgced::celltype_3 -offvalue 0 -onvalue 8
		pack $w.f1.c3 -fill x
		checkbutton $w.f1.c4 -text "Angle block (16)" -anchor w \
				-borderwidth 0 -variable ::pgced::celltype_4 -offvalue 0 -onvalue 16
		pack $w.f1.c4 -fill x

		frame $w.ft -borderwidth 2 -relief groove
		pack  $w.ft -fill x -padx 4

		label $w.ft.lab -text "Angle type:" -anchor w
		pack  $w.ft.lab -fill x
		radiobutton $w.ft.r0 -variable ::pgced::celltype_67 -value 0 -anchor w \
				-text "normal (0)" -pady 0 -bd 0
		pack  $w.ft.r0 -fill x
		radiobutton $w.ft.r1 -variable ::pgced::celltype_67 -value 64 -anchor w \
				-text "first of angle block (64)" -pady 0 -bd 0
		pack  $w.ft.r1 -fill x
		radiobutton $w.ft.r2 -variable ::pgced::celltype_67 -value 128 -anchor w \
				-text "middle of angle block (128)" -pady 0 -bd 0
		pack  $w.ft.r2 -fill x
		radiobutton $w.ft.r3 -variable ::pgced::celltype_67 -value 192 -anchor w \
				-text "last of angle block (192)" -pady 0 -bd 0
		pack  $w.ft.r3 -fill x

		frame  $w.fb
		pack   $w.fb -anchor e -padx 4 -pady 6
		button $w.fb.ok -text "OK" -width 10 -command {set ::pgced::celltypedone 1}
		pack   $w.fb.ok -side left -padx 4
		button $w.fb.cancel -text "Cancel" -width 10 -command {set ::pgced::celltypedone 0}
		pack   $w.fb.cancel -side left

		update
		set wx [expr [winfo rootx .pgceditor]+(([winfo width  .pgceditor]-[winfo reqwidth  $w])/2)]
		set wy [expr [winfo rooty .pgceditor]+(([winfo height .pgceditor]-[winfo reqheight $w])/2)-25]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}

	set ct [$parentbutton cget -text]
	set ::pgced::celltype_0  [expr $ct & 1]
	set ::pgced::celltype_1  [expr $ct & 2]
	set ::pgced::celltype_2  [expr $ct & 4]
	set ::pgced::celltype_3  [expr $ct & 8]
	set ::pgced::celltype_4  [expr $ct & 16]
	set ::pgced::celltype_67 [expr $ct & 192]

	wm deiconify $w
	grab set $w

	set ::pgced::celltypedone -1
	tkwait variable ::pgced::celltypedone
	grab release $w
	grab set .pgceditor
	wm withdraw $w
	if $::pgced::celltypedone {
		set  r $::pgced::celltype_0
		incr r $::pgced::celltype_1
		incr r $::pgced::celltype_2
		incr r $::pgced::celltype_3
		incr r $::pgced::celltype_4
		incr r $::pgced::celltype_67
		$parentbutton configure -text $r
		if {$::pgced::celltype_3 == 8} {
			set $layerbreakvar 0
		} else {
			set $layerbreakvar 1
		}

		return $r
	} else {
		return ""
	}
}

proc set_layer_break {breakflag typebutton} {
	set typeval [$typebutton cget -text]
	if [subst $$breakflag] {
		set typeval [expr $typeval & 0XF7]
	} else {
		set typeval [expr $typeval | 8]
	}
	$typebutton configure -text $typeval
}


proc reassign_VOB_ID {parent cur_cellnum cur_vobid cur_cellid cur_numprograms cur_numcells} {
	set cur_type  [lindex $::editor_cmds_from 0]
	set cur_index [lindex $::editor_cmds_from 1]
	set cur_vts   [lindex [split $cur_index ","] 0]

	# Search for all VOB IDs in the current VTS, and for all Cell IDs in each VOB IDs
	# and build the list of vobids, and the lists of cellids for all vobids.
	# Also, for each pair VOB/Cell ID, keep track of the belonging PGC's cell.
	set all_vobids ""
	if {$cur_type == "menu"} {
		for {set lu 1} {$lu <= $::menus($cur_vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($cur_vts,$lu,numpgcs)} {incr pgc} {
				set numcells $::menus($cur_vts,$lu,$pgc,numcells)
				set cellpostable $::menus($cur_vts,$lu,$pgc,cellpostable)
				for {set cell 1} {$cell <= $numcells} {incr cell} {
					set cellidx [expr ($cell - 1) * 4]
					set vobid  [get_int cellpostable $cellidx 2]
					set cellid [get_int cellpostable [expr $cellidx + 3] 1]
					if {[lsearch $all_vobids $vobid] == -1} {
						lappend all_vobids $vobid
						set all_cellids($vobid) ""
					}
					if {[lsearch $all_cellids($vobid) $cellid] == -1} {
						lappend all_cellids($vobid) $cellid
					}
					set src_pgc_cell($vobid,$cellid) [list "menu" $lu $pgc $cell]
				}
			}
		}
	} else {
		for {set pgc 1} {$pgc <= $::pgcs($cur_vts,numpgcs)} {incr pgc} {
			set cellpostable $::pgcs($cur_vts,$pgc,cellpostable)
			set numcells $::pgcs($cur_vts,$pgc,numcells)
			for {set cell 1} {$cell <= $numcells} {incr cell} {
				set cellidx [expr ($cell - 1) * 4]
				set vobid  [get_int cellpostable $cellidx 2]
				set cellid [get_int cellpostable [expr $cellidx + 3] 1]
				if {[lsearch $all_vobids $vobid] == -1} {
					lappend all_vobids $vobid
					set all_cellids($vobid) ""
				}
				if {[lsearch $all_cellids($vobid) $cellid] == -1} {
					lappend all_cellids($vobid) $cellid
				}
				set src_pgc_cell($vobid,$cellid) [list "pgc" 0 $pgc $cell]
			}
		}
	}
	set all_vobids [lsort -integer $all_vobids]
	foreach vobid $all_vobids {
		set all_cellids($vobid) [lsort -integer $all_cellids($vobid)]
	}

	# GUI
	set w ".assignvobid"
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm iconname $w Dialog
	wm protocol $w WM_DELETE_WINDOW {set ::pgced::vobiddone 0}
	bind $w <Escape> "set ::pgced::vobiddone 0"
	wm transient $w $parent
	if {$::tcl_platform(platform) == "windows"} {
		catch {wm attributes $w -toolwindow 1}
	}
	wm title $w "Assign VOB/Cell ID for PGC Cell $cur_cellnum"

	set txt "Here, you can assing another VOB ID and/or Cell ID"
	set txt "$txt\nto the Cell $cur_cellnum of the current PGC.\n"
	set txt "$txt\nWarning: The old VOB/Cell ID will be unavailable"
	set txt "$txt\nin the future, unless it is reused in another cell !"
	label $w.lab -text $txt -justify left
	pack  $w.lab -fill x -padx 2

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -fill x -padx 4
	frame $w.f1.l1
	pack  $w.f1.l1 -fill x
	label $w.f1.l1.lab -text "New VOB ID:" -width 17 -anchor w
	pack  $w.f1.l1.lab -side left
	entry $w.f1.l1.e -width 3 -state readonly -justify right -bd 0 -relief flat \
			-textvariable ::pgced::new_vobid
	pack  $w.f1.l1.e -side left
	scale $w.f1.s1 -showvalue false -from 0 -to [expr [llength $all_vobids]-1] \
			-orient horizontal -variable ::pgced::vobid_scale \
			-command "::pgced::validate_vobid_scale $w \{$all_vobids\} all_cellids"
	pack  $w.f1.s1 -padx 1 -fill x

	frame $w.f1.l2
	pack  $w.f1.l2 -fill x
	label $w.f1.l2.lab -text "New Cell ID:" -width 17 -anchor w
	pack  $w.f1.l2.lab -side left
	entry $w.f1.l2.e -width 3 -state readonly -justify right -bd 0 -relief flat \
			-textvariable ::pgced::new_cellid
	pack  $w.f1.l2.e -side left
	scale $w.f1.s2 -showvalue false -from 0 -to [expr [llength $all_cellids($cur_vobid)]-1] \
			-orient horizontal -variable ::pgced::cellid_scale \
			-command "::pgced::validate_cellid_scale \{$all_cellids($cur_vobid)\}"
	pack  $w.f1.s2 -padx 1 -fill x

	frame  $w.fb
	pack   $w.fb -anchor e -padx 4 -pady 4
	button $w.fb.ok -text "OK" -width 10 -command {set ::pgced::vobiddone 1}
	pack   $w.fb.ok -side left -padx 4
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::pgced::vobiddone 0}
	pack   $w.fb.cancel -side left

	update
	set wx [expr [winfo rootx $parent]+(([winfo width  $parent]-[winfo reqwidth  $w])/2)]
	set wy [expr [winfo rooty $parent]+(([winfo height $parent]-[winfo reqheight $w])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm resizable $w 0 0

	wm deiconify $w
	grab set $w

	set ::pgced::vobid_scale  [lsearch $all_vobids $cur_vobid]
	set ::pgced::cellid_scale [lsearch $all_cellids($cur_vobid) $cur_cellid]

	set ::pgced::vobiddone -1
	tkwait variable ::pgced::vobiddone
	grab release $w
	if {$parent != "."} {
		grab set $parent
	}
	wm withdraw $w
	if {!$::pgced::vobiddone} return
	set new_vobid  $::pgced::new_vobid
	set new_cellid $::pgced::new_cellid
	if {$new_vobid == $cur_vobid && $new_cellid == $cur_cellid} return

	# OK, we have the new VOB ID in $new_vobid and the new Cell ID in $new_cellid.
	# Begin the job!

	set src $src_pgc_cell($new_vobid,$new_cellid)
	set src_type [lindex $src 0]
	set src_lu   [lindex $src 1]
	set src_pgc  [lindex $src 2]
	set src_cell [lindex $src 3]

	# get the relevant entry in the source cell playback table
	if {$src_type == "menu"} {
		set src_index "$cur_vts,$src_lu,$src_pgc"
	} else {
		set src_index "$cur_vts,$src_pgc"
	}
	set src_cellpbtable [subst \$::${src_type}s($src_index,cellpbtable)]
	set src_cellpbidx   [expr ($src_cell - 1) * 24]
	set src_cellpbentry [string range $src_cellpbtable $src_cellpbidx [expr $src_cellpbidx + 23]]

	set cur_cellpbidx [expr ($cur_cellnum - 1) * 24]

	# copy the cell type and the cell command number from the original cell
	set cur_type [expr [get_int ::pgced::__cellpbtable $cur_cellpbidx 1] | 2] ;# add SRC Discontinuity flag
	set cur_cmdnum [get_int ::pgced::__cellpbtable [expr $cur_cellpbidx + 3] 1]
	set_int src_cellpbentry 0 $cur_type 1
	set_int src_cellpbentry 3 $cur_cmdnum 1

	# replace the new entry in the current cell playback table
	set ::pgced::__cellpbtable "[string range $::pgced::__cellpbtable 0 [expr $cur_cellpbidx-1]]$src_cellpbentry[string range $::pgced::__cellpbtable [expr $cur_cellpbidx+24] end]"

	# and set the VOB ID and Cell ID in the Cell Position table
	set cur_cellposidx [expr ($cur_cellnum - 1) * 4]
	set_int ::pgced::__cellpostable $cur_cellposidx $new_vobid 2
	set_int ::pgced::__cellpostable [expr $cur_cellposidx + 3] $new_cellid 1

	# Set the SCR Discontinuity flag in the NEXT Cell entry
	if {$cur_numcells > $cur_cellnum} {
		set idx [expr $cur_cellnum * 24]
		set_int ::pgced::__cellpbtable $idx [expr [get_int ::pgced::__cellpbtable $idx 1] | 2] 1
	}

	# reload the new tables in the editor
	set total_time [build_cell_table $parent $cur_numprograms $cur_numcells]

	# correct the total playback time
	set ::pgced::__pgcheader "[string range $::pgced::__pgcheader 0 3]$total_time[string range $::pgced::__pgcheader 8 end]"
	set ::pgced::pbtimetxt [BCDtime $total_time true]
}

proc validate_vobid_scale {w all_vobids all_cellids_upvar val} {
	upvar $all_cellids_upvar all_cellids 
	set ::pgced::new_vobid [lindex $all_vobids $val]
	$w.f1.s2 configure -to [expr [llength $all_cellids($::pgced::new_vobid)]-1] \
			-command "::pgced::validate_cellid_scale \{$all_cellids($::pgced::new_vobid)\}"
	set pos [lsearch $all_cellids($::pgced::new_vobid) $::pgced::new_cellid]
	if {$pos == -1} {
		$w.f1.s2 set 0
		set ::pgced::new_cellid [lindex $all_cellids($::pgced::new_vobid) 0]
	} else {
		$w.f1.s2 set $pos
	}
}
proc validate_cellid_scale {all_cellids val} {
	set ::pgced::new_cellid [lindex $all_cellids $val]
}


}	;# end of namespace ::pgced
